import asyncio
import os
import sys
import time
from collections.abc import Callable, Coroutine, Generator, Iterator
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
from typing import Any

import aiohttp
import matplotlib.pyplot as plt
import pyinstrument
import urllib3
from matplotlib.axes import Axes

import httpcore2

PORT = 1234
URL = f"http://localhost:{PORT}/req"
REPEATS = 10
REQUESTS = 500
CONCURRENCY = 20
POOL_LIMIT = 100
PROFILE = False
os.environ["HTTPCORE_PREFER_ANYIO"] = "0"


def duration(start: float) -> int:
    return int((time.monotonic() - start) * 1000)


@contextmanager
def profile() -> Generator[None]:
    if not PROFILE:
        yield
        return
    with pyinstrument.Profiler() as profiler:
        yield
    profiler.open_in_browser()


async def run_async_requests(axis: Axes) -> None:
    async def gather_limited_concurrency(
        coros: Iterator[Coroutine[Any, Any, Any]], concurrency: int = CONCURRENCY
    ) -> None:
        sem = asyncio.Semaphore(concurrency)

        async def coro_with_sem(coro: Coroutine[Any, Any, Any]) -> None:
            async with sem:
                await coro

        await asyncio.gather(*(coro_with_sem(c) for c in coros))

    async def httpcore_get(pool: httpcore2.AsyncConnectionPool, timings: list[int]) -> None:
        start = time.monotonic()
        res = await pool.request("GET", URL)
        assert len(await res.aread()) == 2000
        assert res.status == 200, f"status_code={res.status}"
        timings.append(duration(start))

    async def aiohttp_get(session: aiohttp.ClientSession, timings: list[int]) -> None:
        start = time.monotonic()
        async with session.request("GET", URL) as res:
            assert len(await res.read()) == 2000
            assert res.status == 200, f"status={res.status}"
        timings.append(duration(start))

    async with httpcore2.AsyncConnectionPool(max_connections=POOL_LIMIT) as pool:
        # warmup
        await gather_limited_concurrency((httpcore_get(pool, []) for _ in range(REQUESTS)), CONCURRENCY * 2)

        timings: list[int] = []
        start = time.monotonic()
        with profile():
            for _ in range(REPEATS):
                await gather_limited_concurrency(httpcore_get(pool, timings) for _ in range(REQUESTS))
        axis.plot([*range(len(timings))], timings, label=f"httpcore (tot={duration(start)}ms)")

    connector = aiohttp.TCPConnector(limit=POOL_LIMIT)
    async with aiohttp.ClientSession(connector=connector) as session:
        # warmup
        await gather_limited_concurrency((aiohttp_get(session, []) for _ in range(REQUESTS)), CONCURRENCY * 2)

        timings = []
        start = time.monotonic()
        for _ in range(REPEATS):
            await gather_limited_concurrency(aiohttp_get(session, timings) for _ in range(REQUESTS))
        axis.plot([*range(len(timings))], timings, label=f"aiohttp (tot={duration(start)}ms)")


def run_sync_requests(axis: Axes) -> None:
    def run_in_executor(fns: Iterator[Callable[[], None]], executor: ThreadPoolExecutor) -> None:
        futures = [executor.submit(fn) for fn in fns]
        for future in futures:
            future.result()

    def httpcore_get(pool: httpcore2.ConnectionPool, timings: list[int]) -> None:
        start = time.monotonic()
        res = pool.request("GET", URL)
        assert len(res.read()) == 2000
        assert res.status == 200, f"status_code={res.status}"
        timings.append(duration(start))

    def urllib3_get(pool: urllib3.HTTPConnectionPool, timings: list[int]) -> None:
        start = time.monotonic()
        res = pool.request("GET", "/req")
        assert len(res.data) == 2000
        assert res.status == 200, f"status={res.status}"
        timings.append(duration(start))

    with httpcore2.ConnectionPool(max_connections=POOL_LIMIT) as pool:
        # warmup
        with ThreadPoolExecutor(max_workers=CONCURRENCY * 2) as exec:
            run_in_executor(
                (lambda: httpcore_get(pool, []) for _ in range(REQUESTS)),
                exec,
            )

        timings: list[int] = []
        exec = ThreadPoolExecutor(max_workers=CONCURRENCY)
        start = time.monotonic()
        with profile():
            for _ in range(REPEATS):
                run_in_executor((lambda: httpcore_get(pool, timings) for _ in range(REQUESTS)), exec)
        exec.shutdown(wait=True)
        axis.plot([*range(len(timings))], timings, label=f"httpcore (tot={duration(start)}ms)")

    with urllib3.HTTPConnectionPool("localhost", PORT, maxsize=POOL_LIMIT) as urllib3_pool:
        # warmup
        with ThreadPoolExecutor(max_workers=CONCURRENCY * 2) as exec:
            run_in_executor(
                (lambda: urllib3_get(urllib3_pool, []) for _ in range(REQUESTS)),
                exec,
            )

        timings = []
        exec = ThreadPoolExecutor(max_workers=CONCURRENCY)
        start = time.monotonic()
        for _ in range(REPEATS):
            run_in_executor(
                (lambda: urllib3_get(urllib3_pool, timings) for _ in range(REQUESTS)),
                exec,
            )
        exec.shutdown(wait=True)
        axis.plot([*range(len(timings))], timings, label=f"urllib3 (tot={duration(start)}ms)")


def main() -> None:
    mode = sys.argv[1] if len(sys.argv) == 2 else None
    assert mode in ("async", "sync"), "Usage: python client.py <async|sync>"

    _fig, ax = plt.subplots()

    if mode == "async":
        asyncio.run(run_async_requests(ax))
    else:
        run_sync_requests(ax)

    plt.legend(loc="upper left")
    ax.set_xlabel("# request")
    ax.set_ylabel("[ms]")
    plt.show()
    print("DONE", flush=True)


if __name__ == "__main__":
    main()



import asyncio
from collections.abc import Awaitable, Callable
from typing import Any

import uvicorn

PORT = 1234
RESP = b"a" * 2000
SLEEP = 0.01


async def app(
    scope: dict[str, Any],
    receive: Callable[[], Awaitable[dict[str, Any]]],
    send: Callable[[dict[str, Any]], Awaitable[None]],
) -> None:
    assert scope["type"] == "http"
    assert scope["path"] == "/req"
    assert not (await receive()).get("more_body", False)

    await asyncio.sleep(SLEEP)
    await send(
        {
            "type": "http.response.start",
            "status": 200,
            "headers": [[b"content-type", b"text/plain"]],
        }
    )
    await send(
        {
            "type": "http.response.body",
            "body": RESP,
        }
    )


if __name__ == "__main__":
    uvicorn.run(
        app,
        port=PORT,
        log_level="error",
        # Keep warmed up connections alive during the test to have consistent results across test runs.
        # This avoids timing differences with connections getting closed and reopened in the background.
        timeout_keep_alive=100,
    )



from __future__ import annotations

import ssl
import typing

import pytest

import httpx2

if typing.TYPE_CHECKING:
    from conftest import TestServer


def test_load_ssl_config() -> None:
    context = httpx2.create_ssl_context()
    assert context.verify_mode == ssl.VerifyMode.CERT_REQUIRED
    assert context.check_hostname is True


def test_load_ssl_config_verify_non_existing_file() -> None:
    with pytest.raises(IOError):
        context = httpx2.create_ssl_context()
        context.load_verify_locations(cafile="/path/to/nowhere")


def test_load_ssl_config_cert_and_key(cert_pem_file: str, cert_private_key_file: str) -> None:
    context = httpx2.create_ssl_context()
    context.load_cert_chain(cert_pem_file, cert_private_key_file)
    assert context.verify_mode == ssl.VerifyMode.CERT_REQUIRED
    assert context.check_hostname is True


@pytest.mark.parametrize("password", [b"password", "password"])
def test_load_ssl_config_cert_and_encrypted_key(
    cert_pem_file: str, cert_encrypted_private_key_file: str, password: bytes | str
) -> None:
    context = httpx2.create_ssl_context()
    context.load_cert_chain(cert_pem_file, cert_encrypted_private_key_file, password)
    assert context.verify_mode == ssl.VerifyMode.CERT_REQUIRED
    assert context.check_hostname is True


def test_load_ssl_config_cert_and_key_invalid_password(
    cert_pem_file: str, cert_encrypted_private_key_file: str
) -> None:
    with pytest.raises(ssl.SSLError):
        context = httpx2.create_ssl_context()
        context.load_cert_chain(cert_pem_file, cert_encrypted_private_key_file, "password1")


def test_load_ssl_config_cert_without_key_raises(cert_pem_file: str) -> None:
    with pytest.raises(ssl.SSLError):
        context = httpx2.create_ssl_context()
        context.load_cert_chain(cert_pem_file)


def test_load_ssl_config_no_verify() -> None:
    context = httpx2.create_ssl_context(verify=False)
    assert context.verify_mode == ssl.VerifyMode.CERT_NONE
    assert context.check_hostname is False


def test_create_ssl_context_verify_str(cert_pem_file: str) -> None:
    with pytest.warns(DeprecationWarning, match="`verify=<str>` is deprecated"):
        context = httpx2.create_ssl_context(verify=cert_pem_file)
    assert context.verify_mode == ssl.VerifyMode.CERT_REQUIRED


def test_create_ssl_context_verify_str_with_cert_raises(cert_pem_file: str, cert_private_key_file: str) -> None:
    with pytest.raises(TypeError, match="cannot be combined with `cert=...`"):
        httpx2.create_ssl_context(verify=cert_pem_file, cert=(cert_pem_file, cert_private_key_file))


def test_SSLContext_with_get_request(server: TestServer, cert_pem_file: str) -> None:
    context = httpx2.create_ssl_context()
    context.load_verify_locations(cert_pem_file)
    response = httpx2.get(server.url, verify=context)
    assert response.status_code == 200


def test_limits_repr() -> None:
    limits = httpx2.Limits(max_connections=100)
    expected = "Limits(max_connections=100, max_keepalive_connections=None, keepalive_expiry=5.0)"
    assert repr(limits) == expected


def test_limits_eq() -> None:
    limits = httpx2.Limits(max_connections=100)
    assert limits == httpx2.Limits(max_connections=100)


def test_timeout_eq() -> None:
    timeout = httpx2.Timeout(timeout=5.0)
    assert timeout == httpx2.Timeout(timeout=5.0)


def test_timeout_all_parameters_set() -> None:
    timeout = httpx2.Timeout(connect=5.0, read=5.0, write=5.0, pool=5.0)
    assert timeout == httpx2.Timeout(timeout=5.0)


def test_timeout_from_nothing() -> None:
    timeout = httpx2.Timeout(None)
    assert timeout.connect is None
    assert timeout.read is None
    assert timeout.write is None
    assert timeout.pool is None


def test_timeout_from_none() -> None:
    timeout = httpx2.Timeout(timeout=None)
    assert timeout == httpx2.Timeout(None)


def test_timeout_from_one_none_value() -> None:
    timeout = httpx2.Timeout(None, read=None)
    assert timeout == httpx2.Timeout(None)


def test_timeout_from_one_value() -> None:
    timeout = httpx2.Timeout(None, read=5.0)
    assert timeout == httpx2.Timeout(timeout=(None, 5.0, None, None))


def test_timeout_from_one_value_and_default() -> None:
    timeout = httpx2.Timeout(5.0, pool=60.0)
    assert timeout == httpx2.Timeout(timeout=(5.0, 5.0, 5.0, 60.0))


def test_timeout_missing_default() -> None:
    with pytest.raises(ValueError):
        httpx2.Timeout(pool=60.0)


def test_timeout_from_tuple() -> None:
    timeout = httpx2.Timeout(timeout=(5.0, 5.0, 5.0, 5.0))
    assert timeout == httpx2.Timeout(timeout=5.0)


def test_timeout_from_config_instance() -> None:
    timeout = httpx2.Timeout(timeout=5.0)
    assert httpx2.Timeout(timeout) == httpx2.Timeout(timeout=5.0)


def test_timeout_repr() -> None:
    timeout = httpx2.Timeout(timeout=5.0)
    assert repr(timeout) == "Timeout(timeout=5.0)"

    timeout = httpx2.Timeout(None, read=5.0)
    assert repr(timeout) == "Timeout(connect=None, read=5.0, write=None, pool=None)"


def test_proxy_from_url() -> None:
    proxy = httpx2.Proxy("https://example.com")

    assert str(proxy.url) == "https://example.com"
    assert proxy.auth is None
    assert proxy.headers == {}
    assert repr(proxy) == "Proxy('https://example.com')"


def test_proxy_with_auth_from_url() -> None:
    proxy = httpx2.Proxy("https://username:password@example.com")

    assert str(proxy.url) == "https://example.com"
    assert proxy.auth == ("username", "password")
    assert proxy.headers == {}
    assert repr(proxy) == "Proxy('https://example.com', auth=('username', '********'))"


def test_invalid_proxy_scheme() -> None:
    with pytest.raises(ValueError):
        httpx2.Proxy("invalid://example.com")



from __future__ import annotations

import typing

import pytest

import httpx2

if typing.TYPE_CHECKING:
    from conftest import TestServer


def test_get(server: TestServer) -> None:
    response = httpx2.get(server.url)
    assert response.status_code == 200
    assert response.reason_phrase == "OK"
    assert response.text == "Hello, world!"
    assert response.http_version == "HTTP/1.1"


def test_post(server: TestServer) -> None:
    response = httpx2.post(server.url, content=b"Hello, world!")
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_post_byte_iterator(server: TestServer) -> None:
    def data() -> typing.Iterator[bytes]:
        yield b"Hello"
        yield b", "
        yield b"world!"

    response = httpx2.post(server.url, content=data())
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_post_byte_stream(server: TestServer) -> None:
    class Data(httpx2.SyncByteStream):
        def __iter__(self) -> typing.Iterator[bytes]:
            yield b"Hello"
            yield b", "
            yield b"world!"

    response = httpx2.post(server.url, content=Data())
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_options(server: TestServer) -> None:
    response = httpx2.options(server.url)
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_head(server: TestServer) -> None:
    response = httpx2.head(server.url)
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_put(server: TestServer) -> None:
    response = httpx2.put(server.url, content=b"Hello, world!")
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_patch(server: TestServer) -> None:
    response = httpx2.patch(server.url, content=b"Hello, world!")
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_delete(server: TestServer) -> None:
    response = httpx2.delete(server.url)
    assert response.status_code == 200
    assert response.reason_phrase == "OK"


def test_stream(server: TestServer) -> None:
    with httpx2.stream("GET", server.url) as response:
        response.read()

    assert response.status_code == 200
    assert response.reason_phrase == "OK"
    assert response.text == "Hello, world!"
    assert response.http_version == "HTTP/1.1"


def test_get_invalid_url() -> None:
    with pytest.raises(httpx2.UnsupportedProtocol):
        httpx2.get("invalid://example.org")


# check that httpcore isn't imported until we do a request
def test_httpcore_lazy_loading(server: TestServer) -> None:
    import sys

    # unload our module if it is already loaded
    if "httpx2" in sys.modules:
        del sys.modules["httpx2"]
        del sys.modules["httpcore2"]
    import httpx2

    assert "httpcore2" not in sys.modules
    _response = httpx2.get(server.url)
    assert "httpcore2" in sys.modules



"""
Server-sent events support, derived from httpx-sse (https://github.com/florimondmanca/httpx-sse).

Copyright (c) 2022 Florimond Manca, MIT License (https://github.com/florimondmanca/httpx-sse/blob/master/LICENSE).
"""

from __future__ import annotations

import json as jsonlib
from collections.abc import AsyncIterator, Iterator
from dataclasses import dataclass

from ._exceptions import TransportError
from ._models import Response

__all__ = ["EventSource", "SSEError", "ServerSentEvent"]


class SSEError(TransportError):
    """
    An error that occurred while connecting to a server-sent events endpoint.
    """


@dataclass(frozen=True)
class ServerSentEvent:
    event: str = "message"
    data: str = ""
    id: str = ""
    retry: int | None = None

    def json(self) -> object:
        return jsonlib.loads(self.data)


class _SSEDecoder:
    def __init__(self) -> None:
        self._event = ""
        self._data: list[str] = []
        self._last_event_id = ""
        self._retry: int | None = None
        self._pending = False

    def decode(self, line: str) -> ServerSentEvent | None:
        if not line:
            if not self._pending:
                return None

            sse = ServerSentEvent(
                event=self._event or "message",
                data="\n".join(self._data),
                id=self._last_event_id,
                retry=self._retry,
            )
            self._event = ""
            self._data = []
            self._retry = None
            self._pending = False
            return sse

        if line.startswith(":"):
            return None

        fieldname, _, value = line.partition(":")
        value = value[1:] if value.startswith(" ") else value

        if fieldname == "event":
            self._event = value
            self._pending = True
        elif fieldname == "data":
            self._data.append(value)
            self._pending = True
        elif fieldname == "id":
            if "\0" not in value:
                self._last_event_id = value
                self._pending = True
        elif fieldname == "retry":
            try:
                self._retry = int(value)
                self._pending = True
            except ValueError:
                pass

        return None


class _SSELineDecoder:
    def __init__(self) -> None:
        self._buffer = ""
        self._trailing_cr = False

    def decode(self, text: str) -> list[str]:
        if self._trailing_cr:
            text = "\r" + text
            self._trailing_cr = False
        if text.endswith("\r"):
            self._trailing_cr = True
            text = text[:-1]

        text = self._buffer + text.replace("\r\n", "\n").replace("\r", "\n")
        lines = text.split("\n")
        self._buffer = lines.pop()
        return lines

    def flush(self) -> list[str]:
        if self._trailing_cr:
            self._buffer += "\n"
            self._trailing_cr = False
        if not self._buffer:
            return []
        lines = self._buffer.split("\n")
        self._buffer = ""
        return lines


class EventSource:
    def __init__(self, response: Response) -> None:
        self._response = response

    @property
    def response(self) -> Response:
        return self._response

    def _check_content_type(self) -> None:
        content_type, _, _ = self._response.headers.get("content-type", "").partition(";")
        if content_type.strip().lower() != "text/event-stream":
            raise SSEError(
                f"Expected response with content type 'text/event-stream', got {content_type.strip()!r}.",
                request=self._response.request,
            )

    def __iter__(self) -> Iterator[ServerSentEvent]:
        self._check_content_type()
        decoder = _SSEDecoder()
        lines = _SSELineDecoder()
        for chunk in self._response.iter_text():
            for line in lines.decode(chunk):
                sse = decoder.decode(line)
                if sse is not None:
                    yield sse
        for line in lines.flush():
            sse = decoder.decode(line)
            if sse is not None:
                yield sse

    async def __aiter__(self) -> AsyncIterator[ServerSentEvent]:
        self._check_content_type()
        decoder = _SSEDecoder()
        lines = _SSELineDecoder()
        async for chunk in self._response.aiter_text():
            for line in lines.decode(chunk):
                sse = decoder.decode(line)
                if sse is not None:
                    yield sse
        for line in lines.flush():
            sse = decoder.decode(line)
            if sse is not None:
                yield sse



import typing

import pytest

import httpcore2

# URL


def test_url() -> None:
    url = httpcore2.URL("https://www.example.com/")
    assert url == httpcore2.URL(scheme="https", host="www.example.com", port=None, target="/")
    assert bytes(url) == b"https://www.example.com/"


def test_url_with_port() -> None:
    url = httpcore2.URL("https://www.example.com:443/")
    assert url == httpcore2.URL(scheme="https", host="www.example.com", port=443, target="/")
    assert bytes(url) == b"https://www.example.com:443/"


def test_url_with_invalid_argument() -> None:
    with pytest.raises(TypeError) as exc_info:
        httpcore2.URL(123)  # type: ignore
    assert str(exc_info.value) == "url must be bytes or str, but got int."


def test_url_cannot_include_unicode_strings() -> None:
    """
    URLs instantiated with strings outside of the plain ASCII range are disallowed,
    but the explicit style allows for these ambiguous cases to be precisely expressed.
    """
    with pytest.raises(TypeError) as exc_info:
        httpcore2.URL("https://www.example.com/☺")
    assert str(exc_info.value) == "url strings may not include unicode characters."

    httpcore2.URL(scheme=b"https", host=b"www.example.com", target="/☺".encode())


def test_url_origin_socks5() -> None:
    url = httpcore2.URL("socks5://127.0.0.1")
    origin = url.origin
    assert origin == httpcore2.Origin(scheme=b"socks5", host=b"127.0.0.1", port=1080)
    assert str(origin) == "socks5://127.0.0.1:1080"

    url = httpcore2.URL("socks5h://127.0.0.1")
    origin = url.origin
    assert origin == httpcore2.Origin(scheme=b"socks5h", host=b"127.0.0.1", port=1080)
    assert str(origin) == "socks5h://127.0.0.1:1080"


# Request


def test_request() -> None:
    request = httpcore2.Request("GET", "https://www.example.com/")
    assert request.method == b"GET"
    assert request.url == httpcore2.URL("https://www.example.com/")
    assert request.headers == []
    assert request.extensions == {}
    assert repr(request) == "<Request [b'GET']>"
    assert repr(request.url) == "URL(scheme=b'https', host=b'www.example.com', port=None, target=b'/')"
    assert repr(request.stream) == "<ByteStream [0 bytes]>"


def test_request_with_target_extension() -> None:
    extensions = {"target": b"/another_path"}
    request = httpcore2.Request("GET", "https://www.example.com/path", extensions=extensions)
    assert request.url.target == b"/another_path"

    extensions = {"target": b"/unescaped|path"}
    request = httpcore2.Request("GET", "https://www.example.com/path", extensions=extensions)
    assert request.url.target == b"/unescaped|path"


def test_request_with_invalid_method() -> None:
    with pytest.raises(TypeError) as exc_info:
        httpcore2.Request(123, "https://www.example.com/")  # type: ignore
    assert str(exc_info.value) == "method must be bytes or str, but got int."


def test_request_with_invalid_url() -> None:
    with pytest.raises(TypeError) as exc_info:
        httpcore2.Request("GET", 123)  # type: ignore
    assert str(exc_info.value) == "url must be a URL, bytes, or str, but got int."


def test_request_with_invalid_headers() -> None:
    with pytest.raises(TypeError) as exc_info:
        httpcore2.Request("GET", "https://www.example.com/", headers=123)  # type: ignore
    assert str(exc_info.value) == "headers must be a mapping or sequence of two-tuples, but got int."


# Response


def test_response() -> None:
    response = httpcore2.Response(200)
    assert response.status == 200
    assert response.headers == []
    assert response.extensions == {}
    assert repr(response) == "<Response [200]>"
    assert repr(response.stream) == "<ByteStream [0 bytes]>"


# Tests for reading and streaming sync byte streams...


class ByteIterator:
    def __init__(self, chunks: list[bytes]) -> None:
        self._chunks = chunks

    def __iter__(self) -> typing.Iterator[bytes]:
        yield from self._chunks


def test_response_sync_read() -> None:
    stream = ByteIterator([b"Hello, ", b"world!"])
    response = httpcore2.Response(200, content=stream)
    assert response.read() == b"Hello, world!"
    assert response.content == b"Hello, world!"


def test_response_sync_streaming() -> None:
    stream = ByteIterator([b"Hello, ", b"world!"])
    response = httpcore2.Response(200, content=stream)
    content = b"".join(list(response.iter_stream()))
    assert content == b"Hello, world!"

    # We streamed the response rather than reading it, so .content is not available.
    with pytest.raises(RuntimeError):
        _ = response.content

    # Once we've streamed the response, we can't access the stream again.
    with pytest.raises(RuntimeError):
        for _chunk in response.iter_stream():
            pass  # pragma: no cover


# Tests for reading and streaming async byte streams...


class AsyncByteIterator:
    def __init__(self, chunks: list[bytes]) -> None:
        self._chunks = chunks

    async def __aiter__(self) -> typing.AsyncIterator[bytes]:
        for chunk in self._chunks:
            yield chunk


@pytest.mark.trio
async def test_response_async_read() -> None:
    stream = AsyncByteIterator([b"Hello, ", b"world!"])
    response = httpcore2.Response(200, content=stream)
    assert await response.aread() == b"Hello, world!"
    assert response.content == b"Hello, world!"


@pytest.mark.trio
async def test_response_async_streaming() -> None:
    stream = AsyncByteIterator([b"Hello, ", b"world!"])
    response = httpcore2.Response(200, content=stream)
    content = b"".join([chunk async for chunk in response.aiter_stream()])
    assert content == b"Hello, world!"

    # We streamed the response rather than reading it, so .content is not available.
    with pytest.raises(RuntimeError):
        _ = response.content

    # Once we've streamed the response, we can't access the stream again.
    with pytest.raises(RuntimeError):
        async for _chunk in response.aiter_stream():
            pass  # pragma: no cover



#!venv/bin/python
import os
import re
import sys
from pprint import pprint

SUBS = [
    (
        "from .._backends.auto import AutoBackend",
        "from .._backends.sync import SyncBackend",
    ),
    ("import trio as concurrency", "from tests.httpcore2 import concurrency"),
    ("AsyncIterator", "Iterator"),
    ("Async([A-Z][A-Za-z0-9_]*)", r"\2"),
    ("async def", "def"),
    ("async with", "with"),
    ("async for", "for"),
    ("await ", ""),
    ("handle_async_request", "handle_request"),
    ("aclose", "close"),
    ("aiter_stream", "iter_stream"),
    ("aread", "read"),
    ("asynccontextmanager", "contextmanager"),
    ("safe_async_iterate", "safe_iterate"),
    ("__aenter__", "__enter__"),
    ("__aexit__", "__exit__"),
    ("__aiter__", "__iter__"),
    ("@pytest.mark.anyio", ""),
    ("@pytest.mark.trio", ""),
    ("AutoBackend", "SyncBackend"),
]
COMPILED_SUBS = [(re.compile(r"(^|\b)" + regex + r"($|\b)"), repl) for regex, repl in SUBS]

USED_SUBS = set()


def unasync_line(line):
    for index, (regex, repl) in enumerate(COMPILED_SUBS):
        old_line = line
        line = re.sub(regex, repl, line)
        if old_line != line:
            USED_SUBS.add(index)
    return line


def unasync_file(in_path, out_path):
    with open(in_path) as in_file:
        with open(out_path, "w", newline="") as out_file:
            for line in in_file.readlines():
                line = unasync_line(line)
                out_file.write(line)


def unasync_file_check(in_path, out_path):
    with open(in_path) as in_file:
        with open(out_path) as out_file:
            for in_line, out_line in zip(in_file.readlines(), out_file.readlines(), strict=True):
                expected = unasync_line(in_line)
                if out_line != expected:
                    print(f"unasync mismatch between {in_path!r} and {out_path!r}")
                    print(f"Async code:         {in_line!r}")
                    print(f"Expected sync code: {expected!r}")
                    print(f"Actual sync code:   {out_line!r}")
                    sys.exit(1)


def unasync_dir(in_dir, out_dir, check_only=False):
    for dirpath, _dirnames, filenames in os.walk(in_dir):
        for filename in filenames:
            if not filename.endswith(".py"):
                continue
            rel_dir = os.path.relpath(dirpath, in_dir)
            in_path = os.path.normpath(os.path.join(in_dir, rel_dir, filename))
            out_path = os.path.normpath(os.path.join(out_dir, rel_dir, filename))
            print(in_path, "->", out_path)
            if check_only:
                unasync_file_check(in_path, out_path)
            else:
                unasync_file(in_path, out_path)


def main():
    check_only = "--check" in sys.argv
    unasync_dir(
        "src/httpcore2/httpcore2/_async",
        "src/httpcore2/httpcore2/_sync",
        check_only=check_only,
    )
    unasync_dir(
        "tests/httpcore2/_async",
        "tests/httpcore2/_sync",
        check_only=check_only,
    )

    if len(USED_SUBS) != len(SUBS):
        unused_subs = [SUBS[i] for i in range(len(SUBS)) if i not in USED_SUBS]

        print("These patterns were not used:")
        pprint(unused_subs)
        exit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

import httpx2


def test_client_base_url() -> None:
    client = httpx2.Client()
    client.base_url = "https://www.example.org/"
    assert isinstance(client.base_url, httpx2.URL)
    assert client.base_url == "https://www.example.org/"


def test_client_base_url_without_trailing_slash() -> None:
    client = httpx2.Client()
    client.base_url = "https://www.example.org/path"
    assert isinstance(client.base_url, httpx2.URL)
    assert client.base_url == "https://www.example.org/path/"


def test_client_base_url_with_trailing_slash() -> None:
    client = httpx2.Client()
    client.base_url = "https://www.example.org/path/"
    assert isinstance(client.base_url, httpx2.URL)
    assert client.base_url == "https://www.example.org/path/"


def test_client_headers() -> None:
    client = httpx2.Client()
    client.headers = {"a": "b"}
    assert isinstance(client.headers, httpx2.Headers)
    assert client.headers["A"] == "b"


def test_client_cookies() -> None:
    client = httpx2.Client()
    client.cookies = {"a": "b"}
    assert isinstance(client.cookies, httpx2.Cookies)
    mycookies = list(client.cookies.jar)
    assert len(mycookies) == 1
    assert mycookies[0].name == "a" and mycookies[0].value == "b"


def test_client_timeout() -> None:
    expected_timeout = 12.0
    client = httpx2.Client()

    client.timeout = expected_timeout

    assert isinstance(client.timeout, httpx2.Timeout)
    assert client.timeout.connect == expected_timeout
    assert client.timeout.read == expected_timeout
    assert client.timeout.write == expected_timeout
    assert client.timeout.pool == expected_timeout


def test_client_event_hooks() -> None:
    def on_request(request: httpx2.Request) -> None:
        pass  # pragma: no cover

    client = httpx2.Client()
    client.event_hooks = {"request": [on_request]}
    assert client.event_hooks == {"request": [on_request], "response": []}


def test_client_trust_env() -> None:
    client = httpx2.Client()
    assert client.trust_env

    client = httpx2.Client(trust_env=False)
    assert not client.trust_env



"""
Type definitions for type checking purposes.
"""

from collections.abc import AsyncIterable, AsyncIterator, Callable, Iterable, Iterator, Mapping, Sequence
from http.cookiejar import CookieJar
from typing import IO, TYPE_CHECKING, Any, Union

if TYPE_CHECKING:
    from ._auth import Auth  # noqa: F401
    from ._config import Proxy, Timeout  # noqa: F401
    from ._models import Cookies, Headers, Request  # noqa: F401
    from ._urls import URL, QueryParams  # noqa: F401


PrimitiveData = str | int | float | bool | None

URLTypes = Union["URL", str]

QueryParamTypes = Union[
    "QueryParams",
    Mapping[str, PrimitiveData | Sequence[PrimitiveData]],
    list[tuple[str, PrimitiveData]],
    tuple[tuple[str, PrimitiveData], ...],
    str,
    bytes,
]

HeaderTypes = Union[
    "Headers",
    Mapping[str, str],
    Mapping[bytes, bytes],
    Sequence[tuple[str, str]],
    Sequence[tuple[bytes, bytes]],
]

CookieTypes = Union["Cookies", CookieJar, dict[str, str], list[tuple[str, str]]]

TimeoutTypes = Union[float | None, tuple[float | None, float | None, float | None, float | None], "Timeout"]
ProxyTypes = Union["URL", str, "Proxy"]
CertTypes = str | tuple[str, str] | tuple[str, str, str]

AuthTypes = Union[tuple[str | bytes, str | bytes], Callable[["Request"], "Request"], "Auth"]

RequestContent = str | bytes | Iterable[bytes] | AsyncIterable[bytes]
ResponseContent = str | bytes | Iterable[bytes] | AsyncIterable[bytes]
ResponseExtensions = Mapping[str, Any]

RequestData = Mapping[str, Any]

FileContent = IO[bytes] | bytes | str
FileTypes = (
    # # file (or bytes)
    FileContent
    # # (filename, file (or bytes))
    | tuple[str | None, FileContent]
    # # (filename, file (or bytes), content_type)
    | tuple[str | None, FileContent, str | None]
    | tuple[str | None, FileContent, str | None, Mapping[str, str]]
)
RequestFiles = Mapping[str, FileTypes] | Sequence[tuple[str, FileTypes]]

RequestExtensions = Mapping[str, Any]

__all__ = ["AsyncByteStream", "SyncByteStream"]


class SyncByteStream:
    def __iter__(self) -> Iterator[bytes]:
        raise NotImplementedError("The '__iter__' method must be implemented.")  # pragma: no cover
        yield b""  # pragma: no cover

    def close(self) -> None:
        """
        Subclasses can override this method to release any network resources
        after a request/response cycle is complete.
        """


class AsyncByteStream:
    async def __aiter__(self) -> AsyncIterator[bytes]:
        raise NotImplementedError("The '__aiter__' method must be implemented.")  # pragma: no cover
        yield b""  # pragma: no cover

    async def aclose(self) -> None:
        pass



from __future__ import annotations

from collections.abc import AsyncIterator, Iterator

import pytest

import httpx2

SSE_BODY = b": this is a comment\nevent: ping\ndata: hello\nid: 1\nretry: 500\n\ndata: first\ndata: second\n\n"


def sse_handler(request: httpx2.Request) -> httpx2.Response:
    return httpx2.Response(200, content=SSE_BODY, headers={"Content-Type": "text/event-stream"})


def test_sse_sync() -> None:
    with httpx2.Client(transport=httpx2.MockTransport(sse_handler)) as client:
        with client.sse("http://testserver/sse") as source:
            events = list(source)

    assert events == [
        httpx2.ServerSentEvent(event="ping", data="hello", id="1", retry=500),
        httpx2.ServerSentEvent(event="message", data="first\nsecond", id="1"),
    ]


@pytest.mark.anyio
async def test_sse_async() -> None:
    async with httpx2.AsyncClient(transport=httpx2.MockTransport(sse_handler)) as client:
        async with client.sse("http://testserver/sse") as source:
            events = [event async for event in source]

    assert events == [
        httpx2.ServerSentEvent(event="ping", data="hello", id="1", retry=500),
        httpx2.ServerSentEvent(event="message", data="first\nsecond", id="1"),
    ]


def test_default_event_is_message() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.event == "message"
    assert event.data == "hi"
    assert event.id == ""
    assert event.retry is None


def test_json() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b'data: {"x": 1}\n\n', headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.json() == {"x": 1}


def test_invalid_retry_is_ignored() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        body = b"retry: not-a-number\ndata: hi\n\n"
        return httpx2.Response(200, content=body, headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.retry is None


def test_id_with_null_is_ignored() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        body = b"id: a\0b\ndata: hi\n\n"
        return httpx2.Response(200, content=body, headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.id == ""


def test_field_without_value() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"data\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.data == ""


def test_last_event_id_persists() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        body = b"id: 1\ndata: a\n\ndata: b\n\n"
        return httpx2.Response(200, content=body, headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            events = list(source)

    assert [event.id for event in events] == ["1", "1"]


def test_blank_lines_after_id_do_not_dispatch() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        body = b"id: 1\ndata: a\n\n\n\ndata: b\n\n"
        return httpx2.Response(200, content=body, headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            events = list(source)

    assert [event.data for event in events] == ["a", "b"]


def test_id_only_block_is_dispatched() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"id: 1\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.id == "1"
    assert event.data == ""


@pytest.mark.parametrize(
    "content_type",
    ["text/event-stream", "text/event-stream; charset=utf-8", "Text/Event-Stream", "TEXT/EVENT-STREAM"],
)
def test_content_type_accepted(content_type: str) -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": content_type})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.data == "hi"


def test_content_type_mismatch_raises() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": "application/json"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            with pytest.raises(httpx2.SSEError, match="text/event-stream"):
                list(source)


@pytest.mark.anyio
async def test_content_type_mismatch_raises_async() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": "application/json"})

    async with httpx2.AsyncClient(transport=httpx2.MockTransport(handler)) as client:
        async with client.sse("http://testserver/sse") as source:
            with pytest.raises(httpx2.SSEError, match="text/event-stream"):
                [event async for event in source]


def test_sets_sse_headers() -> None:
    captured: dict[str, str] = {}

    def handler(request: httpx2.Request) -> httpx2.Response:
        captured.update(request.headers)
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            list(source)

    assert captured["accept"] == "text/event-stream"
    assert captured["cache-control"] == "no-store"


def test_user_headers_take_precedence() -> None:
    captured: dict[str, str] = {}

    def handler(request: httpx2.Request) -> httpx2.Response:
        captured.update(request.headers)
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse", headers={"Accept": "text/event-stream, application/json"}) as source:
            list(source)

    assert captured["accept"] == "text/event-stream, application/json"


def test_response_is_accessible() -> None:
    with httpx2.Client(transport=httpx2.MockTransport(sse_handler)) as client:
        with client.sse("http://testserver/sse") as source:
            assert isinstance(source.response, httpx2.Response)
            assert source.response.status_code == 200


def test_post_method() -> None:
    captured: list[str] = []

    def handler(request: httpx2.Request) -> httpx2.Response:
        captured.append(request.method)
        return httpx2.Response(200, content=b"data: hi\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse", method="POST", json={"q": 1}) as source:
            list(source)

    assert captured == ["POST"]


def test_chunk_boundaries_and_crlf_sync() -> None:
    def chunks() -> Iterator[bytes]:
        yield b"data: he"
        yield b"llo\r\n\r"
        yield b"\ndata: wor"
        yield b"ld\r\n\r\n"

    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=chunks(), headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            events = list(source)

    assert [event.data for event in events] == ["hello", "world"]


@pytest.mark.anyio
async def test_chunk_boundaries_and_crlf_async() -> None:
    async def chunks() -> AsyncIterator[bytes]:
        yield b"data: he"
        yield b"llo\r\n\r"
        yield b"\ndata: wor"
        yield b"ld\r\n\r\n"

    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=chunks(), headers={"Content-Type": "text/event-stream"})

    async with httpx2.AsyncClient(transport=httpx2.MockTransport(handler)) as client:
        async with client.sse("http://testserver/sse") as source:
            events = [event async for event in source]

    assert [event.data for event in events] == ["hello", "world"]


def test_event_without_trailing_blank_line() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"data: hi", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            events = list(source)

    assert events == []


def test_leading_blank_line_is_ignored() -> None:
    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=b"\ndata: hi\n\n", headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.data == "hi"


def test_event_dispatched_at_eof_on_trailing_cr_sync() -> None:
    def chunks() -> Iterator[bytes]:
        yield b"data: hi\n"
        yield b"\r"

    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=chunks(), headers={"Content-Type": "text/event-stream"})

    with httpx2.Client(transport=httpx2.MockTransport(handler)) as client:
        with client.sse("http://testserver/sse") as source:
            (event,) = list(source)

    assert event.data == "hi"


@pytest.mark.anyio
async def test_event_dispatched_at_eof_on_trailing_cr_async() -> None:
    async def chunks() -> AsyncIterator[bytes]:
        yield b"data: hi\n"
        yield b"\r"

    def handler(request: httpx2.Request) -> httpx2.Response:
        return httpx2.Response(200, content=chunks(), headers={"Content-Type": "text/event-stream"})

    async with httpx2.AsyncClient(transport=httpx2.MockTransport(handler)) as client:
        async with client.sse("http://testserver/sse") as source:
            events = [event async for event in source]

    assert [event.data for event in events] == ["hi"]
